#!/usr/bin/python
# -*- coding: utf-8 -*-

from epyk.core.py import primitives
from epyk.core.html.options import Options, OptionsWithTemplates
from epyk.core.js import Imports
from epyk.core.js.packages import packageImport
from epyk.core.js import JsUtils


class OptionExtraKeys(Options):

    def alt_f(self, value):
        """

        :param value: String. The event code reference.
        """
        self._config(value, "'Alt-F'")

    def ctrl_space(self, value):
        self._config(value, "'Ctrl-Space'")

    def f_11(self, js_funcs, profile=None):
        """

        :param js_funcs: List | String. Javascript functions.
        :param profile: Boolean | Dictionary. Optional. A flag to set the component performance storage.
        """
        if not isinstance(js_funcs, list):
            js_funcs = [js_funcs]
        self._config(
            'function(cm) {%s}' % JsUtils.jsConvertFncs(js_funcs, toStr=True, profile=profile), "'F11'", js_type=True)

    def esc(self, js_funcs, profile=None):
        """

        :param js_funcs: List | String. Javascript functions.
        :param profile: Boolean | Dictionary. Optional. A flag to set the component performance storage.
        """
        if not isinstance(js_funcs, list):
            js_funcs = [js_funcs]
        self._config(
            'function(cm) {%s}' % JsUtils.jsConvertFncs(js_funcs, toStr=True, profile=profile), "'Esc'", js_type=True)

    def ctrl_q(self, js_funcs, profile=None):
        """

        :param js_funcs: List | String. Javascript functions.
        :param profile: Boolean | Dictionary. Optional. A flag to set the component performance storage.
        """
        if not isinstance(js_funcs, list):
            js_funcs = [js_funcs]
        self._config(
            'function(cm) {%s}' % JsUtils.jsConvertFncs(js_funcs, toStr=True, profile=profile), "'Ctrl-Q'",
            js_type=True)


class OptionPanels:

    def __init__(self, component: primitives.HtmlModel, page: primitives.PageModel = None):
        self.component, self.page = component, page
        if page is None:
            self.page = component.page

    def add(self, content, html_code, position="top"):
        """
        Add a panel to the codemirror component.

        Related Pages:

          https://codemirror.net/demo/panel.html

        :param content: String. The panel content.
        :param html_code: String. An identifier for this component (on both Python and Javascript side).
        :param position: String. Optional. The position of the panel on the component.
      """
        html_code = JsUtils.jsConvertData(html_code, None)
        content = JsUtils.jsConvertData(content, None)
        position = JsUtils.jsConvertData(position, None)
        return '''
if ((typeof window.editor_panels === 'undefined') || !(%(htmlCode)s in window.editor_panels)){
  var node = document.createElement("div");
  var widget, close, label;
  node.id = %(htmlCode)s;
  node.className = "panel " + %(position)s;
  close = node.appendChild(document.createElement("a"));
  close.setAttribute("class", "remove-panel");
  close.textContent = "✖";
  CodeMirror.on(close, "mousedown", function(e) {
    e.preventDefault(); window.editor_panels[node.id].clear()
    delete window.editor_panels[node.id]});
  label = node.appendChild(document.createElement("span"));
  label.textContent = %(content)s;
  if (typeof window.editor_panels === 'undefined'){window.editor_panels = {}}
  window.editor_panels[node.id] = %(editorId)s.addPanel(node, {position: %(position)s, stable: true});
}''' % {"htmlCode": html_code, "position": position, "content": content, "editorId": self.component.editorId}

    def remove(self, html_code):
        """
        Remove a panel visible on the codemirror component.

        Related Pages:

          https://codemirror.net/demo/panel.html

        :param html_code: String. An identifier for this component (on both Python and Javascript side).
        """
        html_code = JsUtils.jsConvertData(html_code, None)
        return '''window.editor_panels[%(htmlCode)s].clear(); delete window.editor_panels[%(htmlCode)s]''' % {
            "htmlCode": html_code}


class OptionCMAddons:

    def __init__(self, options: Options):
        self.__option = options
        self.component = options.component
        self.page = options.page

    @packageImport('codemirror-search', 'codemirror-search')
    def search(self):
        """
        Adds the getSearchCursor(query, start, options) => cursor method to CodeMirror instances, which can be used to
        implement search/replace functionality. query can be a regular expression or a string.
        start provides the starting position of the search.

        Related Pages:

          https://codemirror.net/demo/search.html
        """
        extra_keys = self.__option._config_sub_data("extraKeys", OptionExtraKeys)
        extra_keys.alt_f("findPersistent")
        return self.component

    @packageImport('codemirror-trailingspace')
    def trailingspace(self):
        """
        Adds an option showTrailingSpace which, when enabled, adds the CSS class cm-trailingspace to stretches of
        whitespace at the end of lines.

        Related Pages:

          https://codemirror.net/demo/trailingspace.html
        """
        self.page.body.style.custom_class({
            "background-image": "url(data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAQAAAACCAYAAAB/qH1jAAAABmJLR0QA/wD/AP+gvaeTAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAB3RJTUUH3QUXCToH00Y1UgAAACFJREFUCNdjPMDBUc/AwNDAAAFMTAwMDA0OP34wQgX/AQBYgwYEx4f9lQAAAABJRU5ErkJggg==)",
            "background-position": "bottom left", "background-repeat": "repeat-x"}, "cm-trailingspace")
        self.__option._config(True, "showTrailingSpace")

    @packageImport('codemirror-fold', 'codemirror-fold')
    def foldcode(self):
        """
        Provides an option foldGutter, which can be used to create a gutter with markers indicating the blocks that can be
        folded.

        Related Pages:

          https://codemirror.net/demo/folding.html
        """
        self.__option._config(True, "foldGutter")
        self.__option._config(True, "lineWrapping")
        extra_keys = self.__option._config_sub_data("extraKeys", OptionExtraKeys)
        extra_keys.ctrl_q("cm.foldCode(cm.getCursor())")
        self.__option.gutters = ["CodeMirror-linenumbers", "CodeMirror-foldgutter"]

    @packageImport('codemirror-highlighter')
    def highlighter(self):
        """
        Adds a highlightSelectionMatches option that can be enabled to highlight all instances of a currently selected word.

        Related Pages:

          https://codemirror.net/demo/matchhighlighter.html
        """
        self.page.body.style.custom_class({"background-color": self.page.theme.colors[1]}, "cm-matchhighlight")
        self.__option._config("{showToken: /\w/, annotateScrollbar: true}", "highlightSelectionMatches", js_type=True)

    @packageImport('codemirror-hint', 'codemirror-hint')
    def hint(self):
        """
        Provides a framework for showing autocompletion hints.

        Related Pages:

          https://codemirror.net/demo/complete.html
        """
        extra_keys = self.__option._config_sub_data("extraKeys", OptionExtraKeys)
        extra_keys.ctrl_space('autocomplete')
        return self.component

    @packageImport('codemirror-fullscreen', 'codemirror-fullscreen')
    def fullscreen(self):
        """
        Defines an option fullScreen that, when set to true, will make the editor full-screen (as in, taking up the whole
        browser window).

        Related Pages:

          https://codemirror.net/demo/fullscreen.html
        """
        extra_keys = self.__option._config_sub_data("extraKeys", OptionExtraKeys)
        extra_keys.f_11('cm.setOption("fullScreen", !cm.getOption("fullScreen"))')
        extra_keys.esc('if (cm.getOption("fullScreen")) cm.setOption("fullScreen", false)')
        return self.component

    @packageImport('codemirror-placeholder')
    def placeholder(self):
        """
        Adds a placeholder option that can be used to make content appear in the editor when it is empty and not focused.

        Related Pages:

          https://codemirror.net/demo/placeholder.html
        """
        return self.__option

    @property
    @packageImport('codemirror-panel')
    def panel(self) -> OptionPanels:
        """ Property to the panel add-on features. """
        return OptionPanels(self.component)


class OptionsHints(Options):

    def hint(self, mappings, js_funcs=None, profile=None):
        """

        Related Pages:

          https://codemirror.net/demo/complete.html
          https://www.codegrepper.com/code-examples/whatever/codemirror+hint+on+every+key

        :param mappings: List of List.
        :param js_funcs: List | String. Optional. Javascript functions.
        """
        if not isinstance(js_funcs, list):
            js_funcs = [js_funcs]
        mappings = JsUtils.jsConvertData(mappings, None)
        self._config('''
function(cm, option){
  return new Promise(function(accept) { 
    setTimeout(function() {
      var cursor = cm.getCursor(), line = cm.getLine(cursor.line);  
      var start = cursor.ch, end = cursor.ch;
      while (start && /\w/.test(line.charAt(start - 1))) --start
      while (end < line.length && /\w/.test(line.charAt(end))) ++end
      var word = line.slice(start, end).toLowerCase(); var comp = %s; %s;
      for (var i = 0; i < comp.length; i++) if (comp[i].indexOf(word) != -1)
      return accept({list: comp[i], from: CodeMirror.Pos(cursor.line, start), to: CodeMirror.Pos(cursor.line, end)})
    }) 
  })
}''' % (mappings, JsUtils.jsConvertFncs(js_funcs, toStr=True, profile=profile)), js_type=True)

    def autocomplete(self):
        """ """
        return '''
      '''

    def snipet(self):
        """

    Related Pages:

      https://codepen.io/_kkeisuke/embed/BJGpqG?height=450&slug-hash=BJGpqG&default-tab=js%2Cresult&embed-version=2&user=_kkeisuke&name=cp_embed_1
    """


class OptionsCode(OptionsWithTemplates):
    component_properties = ('stringify',)

    @property
    def addons(self) -> OptionCMAddons:
        """
        The addon directory in the distribution contains a number of reusable components that implement extra editor
        functionality.

        Related Pages:

          https://codemirror.net/doc/manual.html#addons
        """
        return OptionCMAddons(self)

    @property
    def value(self):
        """
        The starting value of the editor. Can be a string, or a document object.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(True)

    @value.setter
    def value(self, value):
        self._config(value)

    @property
    def mode(self):
        """
        The mode to use. When not given, this will default to the first mode that was loaded.
        It may be a string, which either simply names the mode or is a MIME type associated with the mode.

        add extra addon: /mode/javascript/javascript.js

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get("")

    @mode.setter
    def mode(self, value):
        MAP_MODES = {"html": "htmlmixed", "yaml": "yaml-frontmatter", "json": "javascript"}
        value = MAP_MODES.get(value, value)
        self.page.imports.append_to(
            "codemirror", [{"script": '%s.js' % value, "path": 'codemirror/%%(version)s/mode/%s/' % value}])
        self._config(value)

    @property
    def lineSeparator(self):
        """
        Explicitly set the line separator for the editor.
        By default (value null), the document will be split on CRLFs as well as lone CRs and LFs, and a single LF will be
        used as line separator in all output (such as getValue).
        When a specific string is given, lines will only be split on that string, and output will, by default, use that
        same separator.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get()

    @lineSeparator.setter
    def lineSeparator(self, value):
        self._config(value)

    @property
    def gutters(self):
        """
        Can be used to add extra gutters (beyond or instead of the line number gutter).
        Should be an array of CSS class names or class name / CSS string pairs, each of which defines a width
        (and optionally a background), and which will be used to draw the background of the gutters.
        May include the CodeMirror-linenumbers class, in order to explicitly set the position of the line number gutter
        (it will default to be to the right of all other gutters).
        These class names are the keys passed to setGutterMarker.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get()

    @gutters.setter
    def gutters(self, value):
        self._config(value)

    @property
    def theme(self):
        """
        The theme to style the editor with.
        You must make sure the CSS file defining the corresponding .cm-s-[name] styles is loaded (see the theme directory
        in the distribution).
        The default is "default", for which colors are included in codemirror.css.
        It is possible to use multiple theming classes at once—for example "foo bar" will assign both the cm-s-foo and the
        cm-s-bar classes to the editor.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get()

    @theme.setter
    def theme(self, value):
        self._config(value)

    @property
    def indentUnit(self):
        """
        How many spaces a block (whatever that means in the edited language) should be indented. The default is 2.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(2)

    @indentUnit.setter
    def indentUnit(self, value):
        self._config(value)

    @property
    def foldGutter(self):
        """
        Whether to use the context-sensitive indentation that the mode provides (or just indent the same as the line
        before). Defaults to true.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(True)

    @foldGutter.setter
    def foldGutter(self, value):
        self._config(value)

    @property
    def smartIndent(self):
        """
        Whether to use the context-sensitive indentation that the mode provides (or just indent the same as the line
        before). Defaults to true.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(True)

    @smartIndent.setter
    def smartIndent(self, value):
        self._config(value)

    @property
    def tabSize(self):
        """
        The width of a tab character. Defaults to 4.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(4)

    @tabSize.setter
    def tabSize(self, value):
        self._config(value)

    @property
    def indentWithTabs(self):
        """
        Whether, when indenting, the first N*tabSize spaces should be replaced by N tabs. Default is false.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(False)

    @indentWithTabs.setter
    def indentWithTabs(self, value):
        self._config(value)

    @property
    def electricChars(self):
        """
        Configures whether the editor should re-indent the current line when a character is typed that might change its
        proper indentation (only works if the mode supports indentation).
        Default is true.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(True)

    @electricChars.setter
    def electricChars(self, value):
        self._config(value)

    @property
    def direction(self):
        """
        Flips overall layout and selects base paragraph direction to be left-to-right or right-to-left.
        Default is "ltr". CodeMirror applies the Unicode Bidirectional Algorithm to each line, but does not autodetect base
        direction — it's set to the editor direction for all lines.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get("ltr")

    @direction.setter
    def direction(self, value):
        self._config(value)

    @property
    def rtlMoveVisually(self):
        """
        Determines whether horizontal cursor movement through right-to-left (Arabic, Hebrew) text is visual (pressing the
        left arrow moves the cursor left) or logical (pressing the left arrow moves to the next lower index in the string,
        which is visually right in right-to-left text). The default is false on Windows, and true on other platforms.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(False)

    @rtlMoveVisually.setter
    def rtlMoveVisually(self, value):
        self._config(value)

    @property
    def keyMap(self):
        """
        Configures the key map to use. The default is "default", which is the only key map defined in codemirror.js itself.
        Extra key maps are found in the key map directory. See the section on key maps for more information.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get("default")

    @keyMap.setter
    def keyMap(self, value):
        self._config(value)

    @property
    def lineWrapping(self):
        """
        Whether CodeMirror should scroll or wrap for long lines. Defaults to false (scroll).

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(False)

    @lineWrapping.setter
    def lineWrapping(self, value):
        self._config(value)

    @property
    def lineNumbers(self):
        """
        Whether to show line numbers to the left of the editor.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(True)

    @lineNumbers.setter
    def lineNumbers(self, value):
        self._config(value)

    @property
    def firstLineNumber(self):
        """
        At which number to start counting lines. Default is 1.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(1)

    @firstLineNumber.setter
    def firstLineNumber(self, value):
        self._config(value)

    @property
    def fixedGutter(self):
        """
        Determines whether the gutter scrolls along with the content horizontally (false) or whether it stays fixed during
        horizontal scrolling (true, the default).

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(True)

    @fixedGutter.setter
    def fixedGutter(self, value):
        self._config(value)

    @property
    def scrollbarStyle(self):
        """
        Chooses a scrollbar implementation. The default is "native", showing native scrollbars.
        The core library also provides the "null" style, which completely hides the scrollbars. Addons can implement
        additional scrollbar models.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get("native")

    @scrollbarStyle.setter
    def scrollbarStyle(self, value):
        self._config(value)

    @property
    def coverGutterNextToScrollbar(self):
        """
        When fixedGutter is on, and there is a horizontal scrollbar, by default the gutter will be visible to the left of
        this scrollbar.
        If this option is set to true, it will be covered by an element with class CodeMirror-gutter-filler.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(False)

    @coverGutterNextToScrollbar.setter
    def coverGutterNextToScrollbar(self, value):
        self._config(value)

    @property
    def inputStyle(self):
        """
        Selects the way CodeMirror handles input and focus.
        he core library defines the "textarea" and "contenteditable" input models. On mobile browsers, the default is
        "contenteditable". On desktop browsers, the default is "textarea".

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get("textarea")

    @inputStyle.setter
    def inputStyle(self, value):
        self._config(value)

    @property
    def readOnly(self):
        """
        This disables editing of the editor content by the user. If the special value "nocursor" is given
        (instead of simply true), focusing of the editor is also disallowed.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(False)

    @readOnly.setter
    def readOnly(self, value):
        self._config(value)

    @property
    def screenReaderLabel(self):
        """
        This label is read by the screenreaders when CodeMirror text area is focused. This is helpful for accessibility.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get("")

    @screenReaderLabel.setter
    def screenReaderLabel(self, value):
        self._config(value)

    @property
    def showCursorWhenSelecting(self):
        """
        Whether the cursor should be drawn when a selection is active. Defaults to false.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(False)

    @showCursorWhenSelecting.setter
    def showCursorWhenSelecting(self, value):
        self._config(value)

    @property
    def lineWiseCopyCut(self):
        """
        When enabled, which is the default, doing copy or cut when there is no selection will copy or cut the whole lines
        that have cursors on them.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(True)

    @lineWiseCopyCut.setter
    def lineWiseCopyCut(self, value):
        self._config(value)

    @property
    def pasteLinesPerSelection(self):
        """
        When pasting something from an external source (not from the editor itself), if the number of lines matches
        the number of selection,
        CodeMirror will by default insert one line per selection. You can set this to false to disable that behavior.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(False)

    @pasteLinesPerSelection.setter
    def pasteLinesPerSelection(self, value):
        self._config(value)

    @property
    def selectionsMayTouch(self):
        """
        Determines whether multiple selections are joined as soon as they touch (the default) or only when they overlap
        (true).

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(True)

    @selectionsMayTouch.setter
    def selectionsMayTouch(self, value):
        self._config(value)

    @property
    def historyEventDelay(self):
        """
        The period of inactivity (in milliseconds) that will cause a new history event to be started when typing or
        deleting. Defaults to 1250.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(1250)

    @historyEventDelay.setter
    def historyEventDelay(self, value):
        self._config(value)

    @property
    def tabindex(self):
        """
        The tab index to assign to the editor. If not given, no tab index will be assigned.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get()

    @tabindex.setter
    def tabindex(self, value):
        self._config(value)

    @property
    def autofocus(self):
        """
        Can be used to make CodeMirror focus itself on initialization. Defaults to off.
        When fromTextArea is used, and no explicit value is given for this option, it will be set to true when either the
        source textarea is focused, or it has an autofocus attribute and no other element is focused.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(False)

    @autofocus.setter
    def autofocus(self, value):
        self._config(value)

    @property
    def dragDrop(self):
        """
        Controls whether drag-and-drop is enabled. On by default.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(True)

    @dragDrop.setter
    def dragDrop(self, value):
        self._config(value)

    @property
    def cursorBlinkRate(self):
        """
        Half-period in milliseconds used for cursor blinking.
        The default blink rate is 530ms. By setting this to zero, blinking can be disabled. A negative value hides the
        cursor entirely.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(530)

    @cursorBlinkRate.setter
    def cursorBlinkRate(self, value):
        self._config(value)

    @property
    def cursorScrollMargin(self):
        """
        How much extra space to always keep above and below the cursor when approaching the top or bottom of the visible
        view in a scrollable document. Default is 0.

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(0)

    @cursorScrollMargin.setter
    def cursorScrollMargin(self, value):
        self._config(value)

    @property
    def cursorHeight(self):
        """
        Determines the height of the cursor. Default is 1, meaning it spans the whole height of the line.
        For some fonts (and by some tastes) a smaller height (for example 0.85), which causes the cursor to not reach all
        the way to the bottom of the line, looks better

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(1)

    @cursorHeight.setter
    def cursorHeight(self, value):
        self._config(value)

    @property
    def matchBrackets(self):
        """
        Defines an option matchBrackets which, when set to true or an options object, causes matching brackets to be
        highlighted whenever the cursor is next to them.
        It also adds a method matchBrackets that forces this to happen once, and a method findMatchingBracket that can be
        used to run the bracket-finding algorithm that this uses internally.
        It takes a start position and an optional config object.
        By default, it will find the match to a matchable character either before or after the cursor (preferring the one
        before), but you can control its behavior with these options:

        Add extra addon: addon/selection/active-line.js

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(False)

    @matchBrackets.setter
    def matchBrackets(self, value):
        self.page.imports.append_to(
            "codemirror", [{"script": 'matchbrackets.min.js', "path": 'codemirror/%(version)s/addon/edit/'}])
        self._config(value)

    @property
    def styleActiveLine(self):
        """
        Defines a styleActiveLine option that, when enabled, gives the wrapper of the line that contains the cursor the
        class CodeMirror-activeline, adds a background with the class CodeMirror-activeline-background, and adds the class
        CodeMirror-activeline-gutter to the line's gutter space is enabled.

        Add extra addon: addon/edit/matchbrackets.js

        Related Pages:

          https://codemirror.net/doc/manual.html#config
        """
        return self._config_get(False)

    @styleActiveLine.setter
    def styleActiveLine(self, value):
        self.page.imports.append_to(
            "codemirror", [{"script": 'active-line.min.js', "path": 'codemirror/%(version)s/addon/selection/'}])
        self._config(value)

    @property
    def hintOptions(self) -> OptionsHints:
        """

        Related Pages:

          hhttps://codemirror.net/demo/complete.html
        """
        self.addons.hint()
        return self._config_sub_data("hintOptions", OptionsHints)

    @property
    def stringify(self):
        return self._config_get(True)

    @stringify.setter
    def stringify(self, flag: bool):
        self._config(flag)
